from flags.state import (
    flag_enabled as base_flag_enabled,
    flag_disabled as base_flag_disabled
)


def flag_enabled(flag_name, request):
    """ Check if a flag is enabled for a given request """
    return base_flag_enabled(flag_name, request=request)


def flag_disabled(flag_name, request):
    """ Check if a flag is disabled for a given request """
    return base_flag_disabled(flag_name, request=request)



from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

# This will be maintained by register() as a global dictionary of
# condition_name: [list of functions], so we can have multiple conditions
# for a given name and they all must pass when checking.
CONDITIONS = {}


class RequiredForCondition(Exception):
    """ Raised when a kwarg that is required for a condition is not given """


def register(condition_name, fn=None):
    """ Register a condition to test for flag state. Can be decorator.
    Conditions can be any callable that takes a value and some number of
    required arguments (specified in 'requires') that were passed as kwargs
    when checking the flag state. """
    global CONDITIONS

    # Don't be a decorator, just register
    if fn is None:
        # Be a decorator
        def decorator(fn):
            register(condition_name, fn=fn)
            return fn
        return decorator

    if condition_name not in CONDITIONS:
        CONDITIONS[condition_name] = []

    CONDITIONS[condition_name].append(fn)


def get_conditions():
    """ Return the names of all available conditions """
    return CONDITIONS.keys()


def get_condition(condition_name):
    """ Generator to fetch condition checkers from the registry """
    if condition_name not in CONDITIONS:
        raise StopIteration
    for condition_fn in CONDITIONS[condition_name]:
        yield condition_fn


@register('boolean')
def boolean_condition(condition, **kwargs):
    """ Basic boolean check """
    try:
        if condition.lower() == 'true':
            return True
        return False
    except AttributeError:
        return bool(condition)


@register('user')
def user_condition(username, request=None, **kwargs):
    """ Does request.user match the expected username? """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'user'")

    User = apps.get_model('auth', 'User')
    try:
        return request.user == User.objects.get(username=username)
    except ObjectDoesNotExist:
        return False


@register('anonymous')
def anonymous_condition(boolean_value, request=None, **kwargs):
    """ request.user an anonymous user, true or false based on boolean_value
    """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'anonymous'")

    return bool(boolean_value) == bool(request.user.is_anonymous())


@register('parameter')
def parameter_condition(param_name, request=None, **kwargs):
    """ is the parameter name part of the GET parameters? """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'parameter'")

    return request.GET.get(param_name) == 'True'


@register('path')
def path_condition(path, request=None, **kwargs):
    """ Does the request's path match the given path? """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'path'")

    return request.path.startswith(path)


@register('site')
def site_condition(site_str, request=None, **kwargs):
    """ Does the requests's Wagtail Site match the given site?
    site_str should be 'hostname:port', or 'hostname [default]'. """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'site'")

    Site = apps.get_model('wagtailcore.Site')

    if '[default]' in site_str:
        # Wagtail Sites on the default port have [default] at the end of
        # their str() form.
        site_str = site_str.replace(' [default]', ':80')
    elif ':' not in site_str:
        # Add a default port if one isn't given
        site_str += ':80'

    hostname, port = site_str.split(':')
    try:
        conditional_site = Site.objects.get(hostname=hostname, port=port)
    except ObjectDoesNotExist:
        return False

    try:
        site = Site.find_for_request(request)
    except AttributeError:
        # We can't do anything with this
        return False

    return conditional_site == site



from flags.settings import get_flags


def flag_state(flag_name, **kwargs):
    """ Return the value for the flag by passing kwargs to its conditions """
    try:
        return get_flags()[flag_name].check_state(**kwargs)
    except KeyError:
        return False


def flag_enabled(flag_name, **kwargs):
    """ Check if a flag is enabled by passing kwargs to its conditions. """
    return flag_state(flag_name, **kwargs)


def flag_disabled(flag_name, **kwargs):
    """ Check if a flag is disabled by passing kwargs to its conditions. """
    return not flag_state(flag_name, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flags', '0009_migrate_to_conditional_state'),
    ]

    operations = [
        migrations.AlterField(
            model_name='flagstate',
            name='name',
            field=models.CharField(max_length=64),
        ),
        migrations.AlterUniqueTogether(
            name='flagstate',
            unique_together=set([('name', 'condition', 'value')]),
        ),
        migrations.RemoveField(
            model_name='flagstate',
            name='enabled',
        ),
        migrations.RemoveField(
            model_name='flagstate',
            name='flag',
        ),
        migrations.RemoveField(
            model_name='flagstate',
            name='site',
        ),
        migrations.DeleteModel(
            name='Flag',
        ),
    ]



from django import forms

from flags.conditions import get_conditions
from flags.models import FlagState
from flags.settings import get_flags

FLAGS_CHOICES = [(flag, flag) for flag in get_flags().keys()]
CONDITIONS_CHOICES = [(c, c) for c in get_conditions()]


class FlagStateForm(forms.ModelForm):
    name = forms.ChoiceField(choices=FLAGS_CHOICES,
                             label="Flag",
                             required=True)
    condition = forms.ChoiceField(choices=CONDITIONS_CHOICES,
                                  label="Is enabled when",
                                  required=True)
    value = forms.CharField(label="Is", required=True)

    class Meta:
        model = FlagState
        fields = ('name', 'condition', 'value')



from django.http import HttpRequest
from django.test import TestCase

from wagtail.wagtailcore.models import Site

from flags.template_functions import flag_disabled, flag_enabled


class TemplateFunctionsTestCase(TestCase):
    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.request = HttpRequest()
        self.request.site = self.site

    def test_flag_enabled_true(self):
        self.assertTrue(flag_enabled('FLAG_ENABLED', request=self.request))

    def test_flag_enabled_false(self):
        self.assertFalse(flag_enabled('FLAG_DISABLED', request=self.request))

    def test_flag_disabled_true(self):
        self.assertTrue(flag_disabled('FLAG_DISABLED', request=self.request))

    def test_flag_disabled_false(self):
        self.assertFalse(flag_disabled('FLAG_ENABLED', request=self.request))



from django.conf.urls import url, include
from django.core.urlresolvers import resolve
from django.http import HttpResponse, Http404
from django.test import TestCase, RequestFactory, override_settings

from flags.urls import flagged_url, flagged_urls


def view(request):
    return HttpResponse('view')


def fallback(request):
    return HttpResponse('fallback')


extra_patterns = [
    url(r'^included-url$', view),
    url(r'^included-url-with-fallback$', view),
]
fallback_patterns = [
    url(r'^included-url-with-fallback$', fallback),
    url(r'^other-included-url$', fallback),
]

urlpatterns = [
    flagged_url('FLAGGED_URL', r'^url-true-no-fallback$', view,
                name='some-view', state=True),
    flagged_url('FLAGGED_URL', r'^url-false-no-fallback$', view,
                name='some-view', state=False),
    flagged_url('FLAGGED_URL', r'^url-true-fallback$', view,
                name='some-view', state=True, fallback=fallback),
    flagged_url('FLAGGED_URL', r'^url-false-fallback$', view,
                name='some-view', state=False, fallback=fallback),

    flagged_url('FLAGGED_URL', r'^include/', include(extra_patterns),
                state=True),
    flagged_url('FLAGGED_URL', r'^include-false/', include(extra_patterns),
                state=False),
    flagged_url('FLAGGED_URL', r'^include-fallback/', include(extra_patterns),
                state=True, fallback=fallback),
    flagged_url('FLAGGED_URL', r'^include-false-fallback/',
                include(extra_patterns), state=True, fallback=fallback),
    flagged_url('FLAGGED_URL', r'^include-fallback-include/',
                include(extra_patterns),
                state=True, fallback=include(fallback_patterns)),
]

with flagged_urls('FLAGGED_URL') as url:
    flagged_patterns_true_no_fallback = [
        url(r'^patterns-true-no-fallback$', view, name='some-view'),
    ]
urlpatterns = urlpatterns + flagged_patterns_true_no_fallback

with flagged_urls('FLAGGED_URL', state=False) as url:
    flagged_patterns_false_no_fallback = [
        url(r'^patterns-false-no-fallback$', view, name='some-view'),
    ]
urlpatterns = urlpatterns + flagged_patterns_false_no_fallback

with flagged_urls('FLAGGED_URL', fallback=fallback) as url:
    flagged_patterns_true_fallback = [
        url(r'^patterns-true-fallback$', view, name='some-view'),
    ]
urlpatterns = urlpatterns + flagged_patterns_true_fallback


@override_settings(
    ROOT_URLCONF=__name__,
)
class FlagCheckTestCase(TestCase):
    def setUp(self):
        self.flag_name = 'FLAGGED_URL'
        self.factory = RequestFactory()

    def get_url_response(self, url):
        request = self.factory.get(url)
        resolved_view, args, kwargs = resolve(url)
        response = resolved_view(request)
        return response

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_true_no_fallback(self):
        response = self.get_url_response('/url-true-no-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_true_no_fallback_false(self):
        with self.assertRaises(Http404):
            self.get_url_response('/url-true-no-fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_no_fallback(self):
        response = self.get_url_response('/url-false-no-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_false_no_fallback_true(self):
        with self.assertRaises(Http404):
            self.get_url_response('/url-false-no-fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_true_fallback(self):
        response = self.get_url_response('/url-true-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_true_fallback_false(self):
        response = self.get_url_response('/url-true-fallback')
        self.assertContains(response, 'fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_fallback(self):
        response = self.get_url_response('/url-false-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_false_fallback_false(self):
        response = self.get_url_response('/url-false-fallback')
        self.assertContains(response, 'fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_true_include_true(self):
        response = self.get_url_response('/include/included-url')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_true_include_false(self):
        with self.assertRaises(Http404):
            self.get_url_response('/include/included-url')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_include(self):
        response = self.get_url_response('/include-false/included-url')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_false_include_true(self):
        with self.assertRaises(Http404):
            self.get_url_response('/include-false/included-url')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_include_fallback(self):
        response = self.get_url_response('/include-fallback/included-url')
        self.assertContains(response, 'fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_true_include_fallback_include(self):
        response = self.get_url_response(
            '/include-fallback-include/included-url-with-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_include_fallback_include(self):
        response = self.get_url_response(
            '/include-fallback-include/included-url-with-fallback')
        self.assertContains(response, 'fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_include_fallback_none(self):
        with self.assertRaises(Http404):
            self.get_url_response(
                '/include-fallback-include/included-url')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_url_true_include_fallback_include_nonmatching_url(self):
        with self.assertRaises(Http404):
            self.get_url_response(
                '/include-fallback-include/other-included-url')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_url_false_include_fallback_include_nonmatching_url(self):
        response = self.get_url_response(
            '/include-fallback-include/other-included-url')
        self.assertContains(response, 'fallback')

    def test_flagged_url_not_callable(self):
        with self.assertRaises(TypeError):
            flagged_url('MY_FLAG', r'^my_url/$', 'string')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_urls_cm_true_no_fallback(self):
        response = self.get_url_response('/patterns-true-no-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_urls_cm_true_no_fallback_false(self):
        with self.assertRaises(Http404):
            self.get_url_response('/patterns-true-no-fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_urls_cm_false_no_fallback(self):
        response = self.get_url_response('/patterns-false-no-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_urls_cm_false_no_fallback_true(self):
        with self.assertRaises(Http404):
            self.get_url_response('/patterns-false-no-fallback')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': True}})
    def test_flagged_urls_cm_true_fallback(self):
        response = self.get_url_response('/patterns-true-fallback')
        self.assertContains(response, 'view')

    @override_settings(FLAGS={'FLAGGED_URL': {'boolean': False}})
    def test_flagged_urls_cm_true_fallback_false(self):
        response = self.get_url_response('/patterns-true-fallback')
        self.assertContains(response, 'fallback')



import logging

from importlib import import_module

from django.apps import apps
from django.conf import settings
from django.utils.functional import cached_property

from flags.conditions import get_condition

logger = logging.getLogger(__name__)

# Private set of all flags from sources in FLAG_SOURCES. This is
# populated on AppConfig.ready() by add_flags_from_sources below.
SOURCED_FLAGS = {}


class DuplicateFlagException(Exception):
    """ Raised when flags are defined in multiple places """


class Flag:
    """ A simple wrapper around feature flags and their conditions """

    def __init__(self, name, conditions={}):
        self.name = name
        self.__conditions = conditions

    def __eq__(self, other):
        """ There can be only one feature flag of a given name """
        return other.name == self.name

    @cached_property
    def configured_conditions(self):
        """ Get all flag conditions configured in settings """
        # Get condition callables for our settings-configured conditions
        condition_fns = [(c, fn, v, None)
                         for c, v in self.__conditions.items()
                         for fn in get_condition(c)]
        return condition_fns

    @cached_property
    def dynamic_conditions(self):
        """ Get dynamic flag conditions from models.FlagState """
        # Get condition callables for our dynamic-configured conditions
        FlagState = apps.get_model('flags', 'FlagState')
        condition_fns = [(s.condition, fn, s.value, s)
                         for s in FlagState.objects.filter(name=self.name)
                         for fn in get_condition(s.condition)]
        return condition_fns

    @cached_property
    def conditions(self):
        """ Get all flag conditions """
        return self.configured_conditions + self.dynamic_conditions

    def check_state(self, **kwargs):
        """ Determine this flag's state based on any of its conditions """
        return any(fn(v, **kwargs) for c, fn, v, o in self.conditions)


def add_flags_from_sources(sources=None):
    """ Read flags from sources defined in settings.FLAG_SOURCES.
    FLAG_SOURCES is expected to be a list of Python module with flags
    specified with variable assignment, e.g. MY_FLAG = True """
    global SOURCED_FLAGS
    flags = getattr(settings, 'FLAGS', {})

    if sources is None:
        sources = getattr(settings, 'FLAG_SOURCES', ())

    for source_str in sources:
        source = import_module(source_str)
        for flag in (f for f in dir(source)
                     if f.isupper() and isinstance(getattr(source, f), bool)):
            if flag in flags or flag in SOURCED_FLAGS:
                raise DuplicateFlagException("{} duplicated in {}".format(
                    flag, source_str))

            SOURCED_FLAGS[flag] = getattr(source, flag)


def get_flags(sourced_flags=None):
    """ Get a dictionary of Flag objects for all flags.
    This combines FLAGS from settings with all possible FLAG_SOURCES. """
    flags_spec = getattr(settings, 'FLAGS', {})
    if sourced_flags is None:
        sourced_flags = SOURCED_FLAGS
    flags_spec.update(sourced_flags)

    flags = {name: Flag(name, conditions=conditions)
             for name, conditions in flags_spec.items()}
    return flags



from django.test import TestCase

from flags.forms import FlagStateForm


class FormTestCase(TestCase):

    def test_valid_data(self):
        form = FlagStateForm({
            'name': 'FLAG_ENABLED',
            'condition': 'boolean',
            'value': 'True'
        })
        self.assertTrue(form.is_valid())
        state = form.save()
        self.assertEqual(state.name, 'FLAG_ENABLED')
        self.assertEqual(state.condition, 'boolean')
        self.assertEqual(state.value, 'True')

    def test_blank_data(self):
        form = FlagStateForm({})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'name': ['This field is required.'],
            'condition': ['This field is required.'],
            'value': ['This field is required.'],
        })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flags', '0007_unique_flag_site'),
    ]

    operations = [
        migrations.AddField(
            model_name='flagstate',
            name='condition',
            field=models.CharField(default=b'boolean', max_length=64),
        ),
        migrations.AddField(
            model_name='flagstate',
            name='name',
            field=models.CharField(default='', max_length=64),
        ),
        migrations.AddField(
            model_name='flagstate',
            name='value',
            field=models.CharField(default=b'True', max_length=127),
        ),
    ]



from django.http import Http404
from django.utils.functional import wraps

from flags.state import flag_state


def flag_check(flag_name, state, fallback=None, **fc_kwargs):
    """ Check that a given flag has the given state.
    If the state does not match, perform the fallback. """
    def decorator(func):
        def inner(request, *args, **kwargs):
            enabled = flag_state(flag_name, request=request, **fc_kwargs)

            if ((state and enabled) or (not state and not enabled)):
                return func(request, *args, **kwargs)
            elif fallback is not None:
                return fallback(request, *args, **kwargs)
            else:
                raise Http404

        return wraps(func)(inner)

    return decorator


def flag_required(flag_name, fallback_view=None, pass_if_set=True):
    return flag_check(flag_name, pass_if_set, fallback=fallback_view)



from django.db import models


class FlagState(models.Model):
    name = models.CharField(max_length=64)
    condition = models.CharField(max_length=64, default='boolean')
    value = models.CharField(max_length=127, default='True')

    class Meta:
        app_label = 'flags'
        unique_together = ('name', 'condition', 'value')

    def __str__(self):
        return "{name} is enabled when {condition} is {value}".format(
            name=self.name, condition=self.condition, value=self.value)



from django.http import HttpRequest
from django.test import TestCase

from wagtail.wagtailcore.models import Site

from flags.models import FlagState
from flags.state import (
    flag_state,
    flag_enabled,
    flag_disabled
)


class FlagStateTestCase(TestCase):
    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.request = HttpRequest()
        self.request.site = self.site

    def test_non_existent_flag(self):
        """ Non-existent flags always have a default state of False """
        self.assertFalse(flag_state('FLAG_DOES_NOT_EXIST'))

    def test_flag_state_enabled(self):
        """ Global flags that are enabled should be True """
        self.assertTrue(flag_state('FLAG_ENABLED'))

    def test_flag_state_disabled(self):
        """ Global flags that are disabled should be False """
        self.assertFalse(flag_state('FLAG_DISABLED'))

    def test_flag_state_bool_false_and_db_site_true(self):
        """ Test state of multiple conditions, one 'site' in database """
        FlagState.objects.create(name='FLAG_DISABLED',
                                 condition='site',
                                 value=str(self.site))
        self.assertTrue(flag_state('FLAG_DISABLED', request=self.request))

    def test_flag_state_bool_true_and_db_site_true(self):
        """ Test state of multiple conditions, one 'site' in database """
        FlagState.objects.create(name='FLAG_ENABLED',
                                 condition='site',
                                 value=str(self.site))
        self.assertTrue(flag_state('FLAG_ENABLED', request=self.request))

    def test_flag_state_non_existent_flag_site(self):
        """ Given a site non-existent flags should still be False """
        self.assertFalse(flag_state('FLAG_DOES_NOT_EXIST',
                                    request=self.request))

    def test_flag_state_site_for_other_site(self):
        """ A site flag enabled for another site should be False """
        other_site = Site.objects.create(
            is_default_site=False,
            root_page_id=self.site.root_page_id,
            hostname='other.host'
        )
        FlagState.objects.create(name='DB_FLAG',
                                 condition='site',
                                 value=str(other_site))
        self.assertFalse(flag_state('DB_FLAG', request=self.request))

    def test_flag_state_site_for_multiple_sites(self):
        """ A site flag enabled for two sites should return True for both """
        other_site = Site.objects.create(
            is_default_site=False,
            root_page_id=self.site.root_page_id,
            hostname='other.host'
        )
        FlagState.objects.create(name='DB_FLAG',
                                 condition='site',
                                 value=str(other_site))
        FlagState.objects.create(name='DB_FLAG',
                                 condition='site',
                                 value=str(self.site))
        self.assertTrue(flag_state('DB_FLAG', request=self.request))

    def test_flag_enabled_enabled(self):
        """ Global flags enabled should be True """
        self.assertTrue(flag_enabled('FLAG_ENABLED'))

    def test_flag_enabled_disabled(self):
        """ Global flags disabled should be False """
        self.assertFalse(flag_enabled('FLAG_DISABLED'))

    def test_flag_disabled_global_disabled(self):
        """ Global flags disabled should be True """
        self.assertTrue(flag_disabled('FLAG_DISABLED'))

    def test_flag_disabled_global_enabled(self):
        """ Global flags enabled should be False """
        self.assertFalse(flag_disabled('FLAG_ENABLED'))



from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpRequest, QueryDict
from django.test import TestCase

from wagtail.wagtailcore.models import Site

from flags.conditions import (
    CONDITIONS,
    RequiredForCondition,
    register,
    get_condition,
    boolean_condition,
    user_condition,
    anonymous_condition,
    parameter_condition,
    path_condition,
    site_condition,
)


class ConditionRegistryTestCase(TestCase):

    def test_register_decorator(self):
        fn = lambda conditional_value: True
        register('decorated')(fn)
        self.assertIn('decorated', CONDITIONS)
        self.assertEqual(CONDITIONS['decorated'], [fn])

    def test_register_fn(self):
        fn = lambda conditional_value: True
        register('undecorated', fn=fn)
        self.assertIn('undecorated', CONDITIONS)
        self.assertEqual(CONDITIONS['undecorated'], [fn])

    def test_register_required_kwargs(self):
        pass

    def test_get_condition(self):
        fn = lambda conditional_value: True
        register('gettable', fn=fn)
        self.assertEqual(list(get_condition('gettable')), [fn])

    def test_get_condition_none(self):
        self.assertEqual(list(get_condition('notgettable')), [])


class BooleanConditionTestCase(TestCase):

    def test_boolean_condition_valid(self):
        self.assertTrue(boolean_condition(True))

    def test_boolean_condition_invalid(self):
        self.assertFalse(boolean_condition(False))

    def test_boolean_condition_valid_string(self):
        self.assertTrue(boolean_condition('True'))
        self.assertTrue(boolean_condition('true'))

    def test_boolean_condition_invalid_string(self):
        self.assertFalse(boolean_condition('False'))
        self.assertFalse(boolean_condition('false'))


class UserConditionTestCase(TestCase):

    def setUp(self):
        user = User.objects.create_user(username='testuser',
                                        email='test@user')
        self.request = HttpRequest()
        self.request.user = user

    def test_user_valid(self):
        self.assertTrue(user_condition('testuser', request=self.request))

    def test_user_invalid(self):
        self.assertFalse(user_condition('nottestuser', request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            user_condition('testuser')


class AnonymousConditionTestCase(TestCase):

    def setUp(self):
        self.request = HttpRequest()

    def test_anonymous_valid(self):
        self.request.user = AnonymousUser()
        self.assertTrue(anonymous_condition(True, request=self.request))

    def test_anonymous_invalid(self):
        user = User.objects.create_user(username='notadminuser',
                                        email='test@user')
        self.request.user = user
        self.assertFalse(anonymous_condition(True, request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            anonymous_condition(True)


class ParameterConditionTestCase(TestCase):

    def setUp(self):
        self.request = HttpRequest()

    def test_parameter_condition_valid(self):
        self.request.GET = QueryDict('query_flag=True')
        self.assertTrue(parameter_condition('query_flag',
                                            request=self.request))

    def test_parameter_condition_invalid(self):
        self.request.GET = QueryDict('query_flag=False')
        self.assertFalse(parameter_condition('query_flag',
                                             request=self.request))

    def test_parameter_condition_non_existent(self):
        self.request.GET = QueryDict('not_query_flag=True')
        self.assertFalse(parameter_condition('query_flag',
                                             request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            parameter_condition('query_flag')


class PathConditionTestCase(TestCase):

    def setUp(self):
        self.request = HttpRequest()

    def test_path_condition_valid_exact(self):
        self.request.path = '/my/path'
        self.assertTrue(path_condition('/my/path', request=self.request))

    def test_path_condition_valid_subpath(self):
        self.request.path = '/my/path/to/somewhere'
        self.assertTrue(path_condition('/my/path', request=self.request))

    def test_path_condition_invalid(self):
        self.request.path = '/not/my/path'
        self.assertFalse(path_condition('/my/path', request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            path_condition('/my/path')


class SiteConditionTestCase(TestCase):

    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.request = HttpRequest()
        self.request.site = self.site

    def test_site_valid_string(self):
        self.assertTrue(site_condition('localhost:80', request=self.request))

    def test_site_valid_string_no_port(self):
        self.assertTrue(site_condition('localhost', request=self.request))

    def test_site_valid_string_default_port(self):
        self.assertTrue(site_condition('localhost [default]',
                                       request=self.request))

    def test_site_valid_site(self):
        self.assertTrue(site_condition(str(self.site), request=self.request))

    def test_site_invalid_site(self):
        self.assertFalse(site_condition('non.existent.site',
                                        request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            site_condition('localhost:80')



try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from django.http import Http404, HttpRequest, HttpResponse
from django.test import TestCase

from flags.decorators import flag_check, flag_required


class FlagCheckTestCase(TestCase):
    def setUp(self):
        self.request = HttpRequest()
        self.request.META['SERVER_NAME'] = 'localhost'
        self.request.META['SERVER_PORT'] = 8000

        self.view = Mock(__name__='view')

    def test_decorated_no_flag_exists(self):
        decorated = flag_check('FLAG_DOES_NOT_EXIST', True)(self.view)
        with self.assertRaises(Http404):
            decorated(self.request)
        self.assertEqual(self.view.call_count, 0)

    def test_decorated_flag_disabled(self):
        decorated = flag_check('FLAG_DISABLED', True)(self.view)
        self.assertRaises(Http404, decorated, self.request)
        self.assertEqual(self.view.call_count, 0)

    def test_decorated_flag_enabled(self):
        def view(request):
            return HttpResponse('ok')

        decorated = flag_check('FLAG_ENABLED', True)(view)
        response = decorated(self.request)
        self.assertContains(response, 'ok')

    def test_fallback_view(self):
        def fallback(request):
            return HttpResponse('fallback')

        decorator = flag_check('FLAG_DISABLED', True, fallback=fallback)
        decorated = decorator(self.view)
        response = decorated(self.request)
        self.assertContains(response, 'fallback')

    def test_pass_if_not_set_no_flag_exists(self):
        def view(request):
            return HttpResponse('ok')

        decorated = flag_check('FLAG_DOES_NOT_EXIST', False)(view)
        response = decorated(self.request)
        self.assertContains(response, 'ok')

    def test_pass_if_not_set_disabled(self):
        def view(request):
            return HttpResponse('ok')

        decorated = flag_check('FLAG_DISABLED', False)(view)
        response = decorated(self.request)
        self.assertContains(response, 'ok')

    def test_pass_if_not_set_enabled(self):
        decorated = flag_check('FLAG_ENABLED', False)(self.view)
        self.assertRaises(Http404, decorated, self.request)
        self.assertEqual(self.view.call_count, 0)

    def test_pass_if_not_set_fallback_view(self):
        def fallback(request):
            return HttpResponse('fallback')

        decorator = flag_check(
            'FLAG_ENABLED',
            False,
            fallback=fallback,
        )

        decorated = decorator(self.view)
        response = decorated(self.request)
        self.assertContains(response, 'fallback')

    def test_flag_required(self):
        def view(request):
            return HttpResponse('ok')

        decorated = flag_required('FLAG_ENABLED')(view)
        response = decorated(self.request)
        self.assertContains(response, 'ok')



from django.test import TestCase

from flags.models import FlagState


class FlagStateTestCase(TestCase):
    def test_flag_str(self):
        state = FlagState.objects.create(name='MY_FLAG',
                                         condition='boolean',
                                         value='True')
        self.assertEqual(str(state),
                         'MY_FLAG is enabled when boolean is True')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def forwards(apps, schema_editor):
    """ Populate the new name, condition, and value fields on FlagState from
    old Flag and FlagState fields """
    FlagState = apps.get_model('flags', 'FlagState')

    for state in FlagState.objects.all():
        state.name = state.flag.key
        state.condition = 'site'
        state.value = "{hostname}:{port}".format(
            hostname=state.site.hostname, port=state.site.port)
        state.save()


def backwards(apps, schema_editor):
    """ Populate Flag and FlagState foreign keys from new FlagState fields """
    Flag = apps.get_model('flags', 'Flag')
    FlagState = apps.get_model('flags', 'FlagState')
    Site = apps.get_model('wagtailcore', 'Site')

    for state in FlagState.objects.all():
        flag = Flag.objects.create(key=state.name)
        state.flag = flag

        if ':' not in state.value:
            state.value += ':80'
        hostname, port = state.value.split(':')
        site = Site.objects.get(hostname=hostname, port=port)
        state.site = site

        state.save()


class Migration(migrations.Migration):

    dependencies = [
        ('flags', '0008_add_state_conditions'),
    ]

    operations = [
        migrations.RunPython(forwards, backwards),
    ]



from django.apps import AppConfig

from flags.settings import add_flags_from_sources


class WagtailFlagsConfig(AppConfig):
    name = 'flags'
    verbose_name = 'Wagtail Flags'

    def ready(self):
        add_flags_from_sources()



from contextlib import contextmanager

try:
    from django.urls import (
        RegexURLPattern,
        RegexURLResolver
    )
except ImportError:
    from django.core.urlresolvers import (
        RegexURLPattern,
        RegexURLResolver
    )

from flags.decorators import flag_check


class FlaggedURLResolver(RegexURLResolver):
    def __init__(self, flag_name, regex, urlconf_name,
                 default_kwargs=None, app_name=None, namespace=None,
                 state=True, fallback=None):
        super(FlaggedURLResolver, self).__init__(
            regex, urlconf_name, default_kwargs=default_kwargs,
            app_name=app_name, namespace=namespace)

        self.flag_name = flag_name
        self.state = state
        self.fallback = fallback
        self.fallback_patterns = []
        if isinstance(self.fallback, (list, tuple)):
            urlconf_module, app_name, namespace = self.fallback
            self.fallback_patterns = RegexURLResolver(
                regex, urlconf_module, None,
                app_name=app_name, namespace=namespace
            ).url_patterns

    @property
    def url_patterns(self):
        # First, add our "positively" flagged URLs, where when the flag
        # matches the defined state, the view is served for the pattern
        # and not the fallback.
        url_patterns = []
        for pattern in super(FlaggedURLResolver, self).url_patterns:
            # Get the fallback view, if there is one, and remove it from
            # the list of fallback patterns.
            fallback = self.fallback
            if isinstance(self.fallback, (list, tuple)):
                fallback = next((p.callback for p in self.fallback_patterns
                                 if p.regex == pattern.regex), None)

            flag_decorator = flag_check(self.flag_name, self.state,
                                        fallback=fallback)
            flagged_pattern = RegexURLPattern(
                pattern.regex.pattern, flag_decorator(pattern.callback),
                pattern.default_args, pattern.name)
            url_patterns.append(flagged_pattern)

        # Next, add "negatively" flagged URLs, where the flag does not match
        # the defined state, for any remaining fallback patterns that didn't
        # match other url patterns.
        url_pattern_regexes = [p.regex for p in url_patterns]
        for pattern in (p for p in self.fallback_patterns
                        if p.regex not in url_pattern_regexes):
            flag_decorator = flag_check(self.flag_name, not self.state)
            flagged_pattern = RegexURLPattern(
                pattern.regex.pattern, flag_decorator(pattern.callback),
                pattern.default_args, pattern.name)
            url_patterns.append(flagged_pattern)

        return url_patterns


def flagged_url(flag_name, regex, view, kwargs=None, name=None,
                state=True, fallback=None):
    """ Make a URL depend on the state of a feature flag """
    if callable(view):
        flagged_view = flag_check(flag_name,
                                  state,
                                  fallback=fallback)(view)
        return RegexURLPattern(regex, flagged_view, kwargs, name)

    elif isinstance(view, (list, tuple)):
        urlconf_module, app_name, namespace = view
        return FlaggedURLResolver(
            flag_name, regex, urlconf_module, kwargs,
            app_name=app_name, namespace=namespace,
            state=state, fallback=fallback)

    else:
        raise TypeError('view must be a callable')


@contextmanager
def flagged_urls(flag_name, state=True, fallback=None):
    """ Flag multiple URLs in the same context
    Returns a url()-compatible wrapper for flagged_url() """
    def flagged_url_wrapper(regex, view, kwargs=None, name=None):
        return flagged_url(flag_name, regex, view, kwargs=kwargs, name=name,
                           state=state, fallback=fallback)
    yield flagged_url_wrapper



from django import template

from flags.state import (
    flag_enabled as base_flag_enabled,
    flag_disabled as base_flag_disabled
)


register = template.Library()


# @register.simple_tag(takes_context=True)
@register.assignment_tag(takes_context=True)
def flag_enabled(context, flag_name):
    request = context['request']
    return base_flag_enabled(flag_name, request=request)


# @register.simple_tag(takes_context=True)
@register.assignment_tag(takes_context=True)
def flag_disabled(context, flag_name):
    request = context['request']
    return base_flag_disabled(flag_name, request=request)



try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from django.test import TestCase, override_settings

import flags.settings

from flags.models import FlagState

from flags.settings import (
    DuplicateFlagException,
    Flag,
    add_flags_from_sources,
    get_flags
)

# Test flag for using this module to test
SOURCED_FLAG = True


class FlagTestCase(TestCase):

    def test_eq(self):
        flag1 = Flag('MY_FLAG')
        flag2 = Flag('MY_FLAG')
        self.assertEqual(flag1, flag2)

    def test_configured_conditions(self):
        flag = Flag('MY_FLAG', {'boolean': True})
        # Check the conditions length
        self.assertEqual(len(list(flag.configured_conditions)), 1)

    def test_dynamic_conditions(self):
        # Add a dyanmic (database) condition
        flag = Flag('MY_FLAG')
        FlagState.objects.create(name='MY_FLAG',
                                 condition='parameter',
                                 value='MY_FLAG')
        self.assertEqual(len(list(flag.dynamic_conditions)), 1)

    def test_conditions(self):
        flag = Flag('MY_FLAG', {'boolean': True})
        FlagState.objects.create(name='MY_FLAG',
                                 condition='parameter',
                                 value='MY_FLAG')
        self.assertEqual(len(list(flag.conditions)), 2)

    def test_check_state(self):
        flag = Flag('MY_FLAG', {'boolean': True})
        self.assertTrue(flag.check_state())

    def test_check_state_no_conditions(self):
        flag = Flag('MY_FLAG', {})
        self.assertFalse(flag.check_state())

    def test_check_state_multiple_conditions(self):
        request = Mock(path='/foo')
        flag = Flag('MY_FLAG', {'boolean': False, 'path': '/foo'})
        self.assertTrue(flag.check_state(request=request))


class SettingsTestCase(TestCase):

    def tearDown(self):
        # Reset SOURCED_FLAGS after each test
        flags.settings.SOURCED_FLAGS = {}

    def test_add_flags_from_sources(self):
        add_flags_from_sources(sources=['flags.tests.test_settings'])
        self.assertTrue(flags.settings.SOURCED_FLAGS['SOURCED_FLAG'])

    def test_add_flags_from_sources_non_existent(self):
        with self.assertRaises(ImportError):
            add_flags_from_sources(sources=['non.existent.module'])

    @override_settings(FLAGS={'GLOBAL_FLAG': {}})
    def test_get_flags(self):
        add_flags_from_sources(sources=['flags.tests.test_settings'])
        self.assertIn('GLOBAL_FLAG', get_flags())
        self.assertIn('SOURCED_FLAG', get_flags())

    @override_settings(FLAGS={'SOURCED_FLAG': {}})
    def test_duplicate_global_flags(self):
        with self.assertRaises(DuplicateFlagException):
            add_flags_from_sources(sources=['flags.tests.test_settings'])



from django.http import HttpRequest
from django.template import Context, Template
from django.test import TestCase

from wagtail.wagtailcore.models import Site


class FlagsTemplateTagsTestCase(TestCase):

    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.request = HttpRequest()
        self.request.site = self.site

    def render_template(self, string, context=None):
        context = context or {'request': self.request}
        context = Context(context)
        return Template(string).render(context)

    def test_flag_enabled_disabled(self):
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_enabled "FLAG_DISABLED"  as test_flag %}'
            '{% if test_flag %}'
            'flag enabled'
            '{% else %}'
            'flag disabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag disabled')

    def test_flag_enabled_does_not_exist(self):
        # Disabled can also mean non-existent
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_enabled "FLAG_DOES_NOT_EXIST"  as test_flag %}'
            '{% if test_flag %}'
            'flag enabled'
            '{% else %}'
            'flag disabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag disabled')

    def test_flag_enabled_enabled(self):
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_enabled "FLAG_ENABLED"  as test_flag %}'
            '{% if test_flag %}'
            'flag enabled'
            '{% else %}'
            'flag disabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag enabled')

    def test_flag_disabled_disabled(self):
        # Disabled can also mean non-existent
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_disabled "FLAG_DISABLED"  as test_flag %}'
            '{% if test_flag %}'
            'flag disabled'
            '{% else %}'
            'flag enabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag disabled')

    def test_flag_disabled_does_not_exist(self):
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_disabled "FLAG_DOES_NOT_EXIST"  as test_flag %}'
            '{% if test_flag %}'
            'flag disabled'
            '{% else %}'
            'flag enabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag disabled')

    def test_flag_disabled_enabled(self):
        rendered = self.render_template(
            '{% load feature_flags %}'
            '{% flag_disabled "FLAG_ENABLED"  as test_flag %}'
            '{% if test_flag %}'
            'flag disabled'
            '{% else %}'
            'flag enabled'
            '{% endif %}'
        )
        self.assertEqual(rendered, 'flag enabled')



from collections import OrderedDict

from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect, render
from django.views.generic import TemplateView

from flags.decorators import flag_check
from flags.forms import FlagStateForm
from flags.models import FlagState
from flags.settings import get_flags


def index(request):
    flags = OrderedDict(sorted(get_flags().items(), key=lambda x: x[0]))
    context = {
        'flag_states': FlagState.objects.all(),
        'flags': flags,
    }
    return render(request, 'flagadmin/index.html', context)


def create(request):
    if request.method == 'POST':
        form = FlagStateForm(request.POST)
        if form.is_valid():
            form.save()
            return redirect('flagadmin:list')
    else:
        form = FlagStateForm()

    context = dict(form=form)
    return render(request, 'flagadmin/flags/create.html', context)


def delete(request, state_id):
    flag_state = get_object_or_404(FlagState, pk=state_id)

    if request.method == 'POST':
        flag_state.delete()
        return redirect('flagadmin:list')

    context = dict(state_str=str(flag_state), state_id=flag_state.pk)
    return render(request, 'flagadmin/flags/delete.html', context)


class FlaggedViewMixin(object):
    flag_name = None
    fallback = None
    condition = True

    def dispatch(self, request, *args, **kwargs):
        if self.flag_name is None:
            raise ImproperlyConfigured(
                "FlaggedViewMixin requires a 'flag_name' argument."
            )

        super_dispatch = super(FlaggedViewMixin, self).dispatch

        decorator = flag_check(
            self.flag_name,
            self.condition,
            fallback=self.fallback,
        )

        return decorator(super_dispatch)(request, *args, **kwargs)


class FlaggedTemplateView(FlaggedViewMixin, TemplateView):
    pass
